/* This file is part of test microkernel for OpenRISC 1000. */
/* (C) 2000 Damjan Lampret, lampret@opencores.org */

#include "machine.h"

/*
 * Context is saved to area pointed by pointer in R3. Original
 * R3 is at memory location 0 and task's PC is at memory location 4.
 */
#define SAVEREGS					\
	l.lw	r3,0(r3);				\
	l.sw	4(r3),r1;				\
	l.sw	8(r3),r2;				\
	l.lw	r2,0(r0);	/* saving original r3	*/\
	l.sw	12(r3),r2;				\
	l.sw	16(r3),r4;				\
	l.sw	20(r3),r5;				\
	l.sw	24(r3),r6;				\
	l.sw	28(r3),r7;				\
	l.sw	32(r3),r8;				\
	l.sw	36(r3),r9;				\
	l.sw	40(r3),r10;				\
	l.sw	44(r3),r11;				\
	l.sw	48(r3),r12;				\
	l.sw	52(r3),r13;				\
	l.sw	56(r3),r14;				\
	l.sw	60(r3),r15;				\
	l.lw	r2,4(r0);	/* saving original PC*/	\
	l.sw	0(r3),r2;				\
							\
	l.addi	r4,r0,SPR_CCR;				\
	l.mfsr	r2,r4;					\
	l.sw	64(r3),r2	/* saving CCR */

/*
 * Pointer to context is in R3. All registers are loaded and execution is
 * transfered to the loaded context's task
 */
#define LOADREGS_N_GO				\
	l.lw	r3,0(r3);			\
						\
	l.lw	r2,0(r3);	/* prepare PC */	\
	l.sw	4(r0),r2;			\
						\
	l.lw	r2,64(r3);	/* load CCR */	\
	l.addi	r4,r0,SPR_CCR;			\
	l.mtsr	r4,r2;				\
						\
	l.lw	r2,12(r3);	/* prepare r3*/	\
	l.sw	0(r0),r2;			\
						\
	l.lw	r1,4(r3);			\
	l.lw	r2,8(r3);			\
	l.lw	r4,16(r3);			\
	l.lw	r5,20(r3);			\
	l.lw	r6,24(r3);			\
	l.lw	r7,28(r3);			\
	l.lw	r8,32(r3);			\
	l.lw	r9,36(r3);			\
	l.lw	r10,40(r3);			\
	l.lw	r11,44(r3);			\
	l.lw	r12,48(r3);			\
	l.lw	r13,52(r3);			\
	l.lw	r14,56(r3);			\
	l.lw	r15,60(r3);			\
						\
	l.lw	r3,4(r0);	/* Task's PC */	\
	l.rfe	r3;		/* Call task */	\
	l.lw	r3,0(r0);	/* Task's r3 */

/*
 * Set new PC in saved context 
 */
#define SET_CONTEXTPC(AREA,SUBROUTINE,TMPREG)	\
	l.lw	AREA,0(AREA);			\
	l.movhi	TMPREG,hi(SUBROUTINE);		\
	l.addi	TMPREG,r0,lo(SUBROUTINE);	\
	l.sw	0(AREA),TMPREG;

/*
 * Printf via or1ksim hook 
 */
#if KERNEL_OUTPUT
#define PRINTF(REG,STR)				\
	l.movhi	REG,hi(STR);			\
	l.addi	REG,r0,lo(STR);			\
	l.sys	202
#else
#define PRINTF(REG,STR)
#endif

/*
 * Reset Exception handler
 */
.org 0x100
_reset:
	l.nop
	l.j	_kernel_init
	l.nop

/*
 * Bus Error Exception handler
 */
.org 0x0200
_buserr:
	l.nop
	l.sw	0(r0),r3	/* Save r3 */
	PRINTF(r3, _buserr_str)
_hang:
	l.j	_hang
	l.nop

_buserr_str:
        .ascii  "Bus error exception.\n\000"

/*
 * External Interrupt Exception handler
 */
.org 0x500
_extint:
	l.nop
	l.sw	0(r0),r3	/* Save r3 */
	PRINTF(r3,_extint_str)
	l.addi	r3,r0,SPR_EPCR0	/* EPCR0 */
	l.mfsr	r3,r3		/* Get EPCR */
	l.sw	4(r0),r3	/* and save it */

	/* now save user task context */
	l.movhi	r3,hi(_task_context)
	l.addi	r3,r0,lo(_task_context)
	SAVEREGS

	/* set kernel context's PC to kernel's scheduler */
	l.movhi	r3,hi(_kernel_context)
	l.addi	r3,r0,lo(_kernel_context)
	SET_CONTEXTPC(r3,_kernel_sched,r4)

	/* load kernel context */
	l.movhi	r3,hi(_kernel_context)
	l.addi	r3,r0,lo(_kernel_context)
	LOADREGS_N_GO

_extint_str:
        .ascii  "External interrupt exception.\n\000"

/*
 * System Call Exception handler
 */
.org 0x0c00
_syscall:
	l.nop
	l.sw	0(r0),r3	/* Save r3 */
	PRINTF(r3,_syscall_str)
	l.addi	r3,r0,SPR_EPCR0	/* EPCR0 */
	l.mfsr	r3,r3		/* Get EPCR */
	l.sw	4(r0),r3	/* and save it */

	/* now save user task context */
	l.movhi	r3,hi(_task_context)
	l.addi	r3,r0,lo(_task_context)
	SAVEREGS

	/* set kernel context's PC to kernel's syscall entry */
	l.movhi	r3,hi(_kernel_context)
	l.addi	r3,r0,lo(_kernel_context)
	SET_CONTEXTPC(r3,_kernel_syscall,r4)

	/* load kernel context */
	l.movhi	r3,hi(_kernel_context)
	l.addi	r3,r0,lo(_kernel_context)
	LOADREGS_N_GO

_syscall_str:
        .ascii  "System call exception.\n\000"

/*
 * Switch to a new context pointed by _task_context
 */
.global _dispatch
.align 4
_dispatch:
	/* set IER (and we do not clear SUPV) */
	l.addi	r3,r0,SPR_SR
	l.addi	r4,r0,SPR_SR_IER | SPR_SR_SUPV
	l.mtsr	r3,r4

	/* load user task GPRs and PC */
	l.movhi	r3,hi(_task_context)
	l.addi	r3,r0,lo(_task_context)
	LOADREGS_N_GO
